@@ -41,5 +41,11 @@ extern const OSSL_DISPATCH p11prov_ec_encoder_priv_key_info_pem_functions[];
 extern const OSSL_DISPATCH
     p11prov_ec_edwards_encoder_priv_key_info_pem_functions[];
 extern const OSSL_DISPATCH p11prov_ec_edwards_encoder_text_functions[];
+extern const OSSL_DISPATCH p11prov_mldsa_encoder_pkcs1_der_functions[];
+extern const OSSL_DISPATCH p11prov_mldsa_encoder_pkcs1_pem_functions[];
+extern const OSSL_DISPATCH p11prov_mldsa_encoder_spki_der_functions[];
+extern const OSSL_DISPATCH p11prov_mldsa_encoder_spki_pem_functions[];
+extern const OSSL_DISPATCH p11prov_mldsa_encoder_priv_key_info_pem_functions[];
+extern const OSSL_DISPATCH p11prov_mldsa_encoder_text_functions[];
 
 #endif /* _ENCODER_H */
