@@ -16,8 +16,13 @@ certutil -N -d "${TOKDIR}" -f "${PINFILE}"
 export P11LIB="${SOFTOKNPATH%%/}/libsoftokn3${SHARED_EXT}"
 export NSS_LIB_PARAMS="configDir=${TOKDIR}"
 
-export TOKENLABEL="NSS Certificate DB"
-export TOKENLABELURI="NSS%20Certificate%20DB"
+if [[ "${PKCS11_PROVIDER_FORCE_FIPS_MODE}" = "1" || "$(cat /proc/sys/crypto/fips_enabled)" = "1" ]]; then
+    export TOKENLABEL="NSS FIPS 140-2 Certificate DB"
+    export TOKENLABELURI="NSS%20FIPS%20140-2%20Certificate%20DB"
+else
+    export TOKENLABEL="NSS Certificate DB"
+    export TOKENLABELURI="NSS%20Certificate%20DB"
+fi
 
 export TOKENOPTIONS="${TOKENOPTIONS}\npkcs11-module-quirks = no-operation-state no-allowed-mechanisms"
 export TOKENCONFIGVARS="export NSS_LIB_PARAMS=configDir=${TOKDIR}"
@@ -27,3 +32,5 @@ export TESTPORT="30000"
 # Edward curves are not supported in NSS yet
 export SUPPORT_ED25519=0
 export SUPPORT_ED448=0
+
+export SUPPORT_ML_DSA=0
