@@ -20,4 +20,4 @@ void ossl_err_print(void);
 EVP_PKEY *load_key(const char *uri);
 X509 *load_cert(const char *uri, const UI_METHOD *ui_method, void *ui_data);
 void hexify(char *out, unsigned char *byte, size_t len);
-EVP_PKEY *util_gen_key(const char *label);
+EVP_PKEY *util_gen_key(const char *type, const char *label);
